# pylint: disable=missing-module-docstring

import csv
import warnings
from ast import literal_eval

import numpy as np

from hrosailing.core.computing import convert_apparent_wind_to_true
from hrosailing.core.data import WeightedPoints
from hrosailing.processing import ArithmeticMeanInterpolator, Ball

from ._basepolardiagram import PolarDiagram


class PolarDiagramPointcloud(PolarDiagram):
    """A class to represent, visualize and work with a polar diagram
    given by a point cloud.

    Parameters
    ----------
    points : array_like of shape (n, 3)
        Initial points of the point cloud, given as a sequence of
        points consisting of wind speed, wind angle and
        boat speed.
        Points with negative wind speeds will be ignored.

    apparent_wind : bool, optional
        Specifies if wind data is given in apparent wind.

        If `True`, data will be converted to true wind.

        Defaults to `False`.

    Attributes
    -----------
    wind_speeds (property) : numpy.ndarray
        All unique wind speeds in the point cloud.

    wind_angles (property) : numpy.ndarray
        All unique wind angles in the point cloud.

    boat_speeds (property) :
        All occurring boat speeds in the point cloud
        (including duplicates).

    points (property) :
        Read only version of all points present in the point cloud.

    See also
    ---------
    `PolarDiagram`
    """

    @property
    def default_points(self):
        """
        Returns all points given in the data.

        See also
        -------
        `PolarDiagram.default_points`
        """
        return self.points

    def get_slices(
        self,
        ws=None,
        n_steps=None,
        full_info=False,
        wa_resolution=100,
        range_=1,
        interpolator=None,
        **kwargs,
    ):
        """
        Parameters
        ---------------------
        wa_resolution : int, optional
            The number of wind angles that will be used for estimation if an
            interpolator is given.

            Defaults to 100.

        range_ : int, optional
            If no interpolator is given, a slice consists of all data points
            with a wind speed with an absolute difference of at most `range_`
            to the corresponding wind speed value.

            Defaults to 1

        interpolator : Interpolator, optional
            The interpolation method used to estimate unknown boat speeds.

            Defaults to `None`

        See also
        --------
        `PolarDiagram.get_slices`
        """
        kwargs["wa_resolution"] = wa_resolution
        kwargs["range_"] = range_
        kwargs["interpolator"] = interpolator
        return super().get_slices(ws, n_steps, full_info, **kwargs)

    def ws_to_slices(
        self, ws, wa_resolution=100, range_=1, interpolator=None, **kwargs
    ):
        """
        See also
        -------
        `PolarDiagramPointcloud.get_slices`
        `PolarDiagram.ws_to_slices`
        """
        all_ws = self.points[:, 0]
        if interpolator is not None:
            default_wind_angles = np.linspace(0, 360, wa_resolution)

        slices = []
        for ws_ in ws:
            if interpolator is None:
                slicing = np.where(np.abs(all_ws - ws_) <= range_)
                slices.append(np.atleast_2d(self.points[slicing]).T)
                continue
            slices.append(self(ws_, default_wind_angles))
        return slices

    def __init__(self, points, apparent_wind=False):
        if apparent_wind:
            points = convert_apparent_wind_to_true(points)
        else:
            points = np.asarray_chkfinite(points)
            points = points[np.where(points[:, 0] >= 0)]
            points[:, 1] %= 360

        self._points = points

    def __str__(self):
        table = ["   TWS      TWA     BSP\n", "++++++  +++++++  ++++++\n"]
        for point in self.points:
            for i in range(3):
                entry = f"{float(point[i]):.2f}"
                if i == 1:
                    table.append(entry.rjust(7))
                    table.append("  ")
                    continue

                table.append(entry.rjust(6))
                table.append("  ")
            table.append("\n")
        return "".join(table)

    def __repr__(self):
        return f"PolarDiagramPointcloud(pts={self.points})"

    def __call__(
        self,
        ws,
        wa,
        interpolator=ArithmeticMeanInterpolator(50),
        neighbourhood=Ball(radius=1),
    ):
        """Returns the value of the polar diagram at a given `ws-wa` point.

        If the `ws-wa` point is in the cloud, the corresponding boat speed is
        returned, otherwise the value is interpolated.

        Parameters
        ----------
        ws : scalar
            Wind speed.

        wa : scalar
            Wind angle.

        interpolator : Interpolator, optional
            Interpolator subclass that determines the interpolation
            method used to determine the value at the `ws-wa` point.

            Defaults to `ArithmeticMeanInterpolator(50)`.

        neighbourhood : Neighbourhood, optional
            Neighbourhood subclass used to determine the points in
            the point cloud that will be used in the interpolation.

            Defaults to `Ball(radius=1)`.

        Returns
        -------
        bsp : scalar
            Boat speed value as determined above.
        """
        if np.any((ws <= 0)):
            raise ValueError("`ws` is non-positive")

        wa %= 360

        cloud = self.points
        point = cloud[np.logical_and(cloud[:, 0] == ws, cloud[:, 1] == wa)]
        if point.size:
            return point[2]

        point = np.array([ws, wa])
        weighted_points = WeightedPoints(data=cloud, weights=1)

        considered_points = neighbourhood.is_contained_in(cloud[:, :2] - point)

        return interpolator.interpolate(
            weighted_points[considered_points], point
        )

    @property
    def default_slices(self):
        ws = self.points[:, 0]
        return np.linspace(ws.min(), ws.max(), 20)

    @property
    def wind_speeds(self):
        """Returns all unique wind speeds in the point cloud."""
        return np.array(sorted(list(set(self.points[:, 0]))))

    @property
    def wind_angles(self):
        """Returns all unique wind angles in the point cloud."""
        return np.array(sorted(list(set(self.points[:, 1]))))

    @property
    def boat_speeds(self):
        """Returns all occurring boat speeds in the point cloud
        (including duplicates).
        """
        return self.points[:, 2]

    @property
    def points(self):
        """Returns a read only version of `self._points`."""
        return self._points.copy()

    def to_csv(self, csv_path):
        """Creates a .csv file with delimiter ',' and the
        following format:

            `PolarDiagramPointcloud`
            TWS,TWA,BSP
            `self.points`

        Parameters
        ----------
        csv_path : path-like
            Path to a .csv-file or where a new .csv file will be created.
        """
        with open(csv_path, "w", newline="", encoding="utf-8") as file:
            csv_writer = csv.writer(file, delimiter=",")
            csv_writer.writerow([self.__class__.__name__])
            csv_writer.writerow(["TWS", "TWA", "BSP"])
            csv_writer.writerows(self.points)

    @classmethod
    def __from_csv__(cls, file):
        csv_reader = csv.reader(file, delimiter=",")
        next(csv_reader)
        points = np.array(
            [[literal_eval(point) for point in row] for row in csv_reader]
        )

        return PolarDiagramPointcloud(points)

    def symmetrize(self):
        """Constructs a symmetric version of the polar diagram,
        by mirroring it at the 0° - 180° axis and returning a new instance.

        Warning
        -------
        Should only be used if all the wind angles of the initial
        polar diagram are on one side of the 0° - 180° axis,
        otherwise this can result in the construction of duplicate points,
        that might overwrite or live alongside old points.
        """
        if not self.points.size:
            return self

        below_180 = [wa for wa in self.wind_angles if wa <= 180]
        above_180 = [wa for wa in self.wind_angles if wa > 180]
        if below_180 and above_180:
            warnings.warn(
                "there are wind angles on both sides of the 0° - 180° axis. "
                "This might result in duplicate data, "
                "which can overwrite or live alongside old data"
            )

        mirrored_points = self.points
        mirrored_points[:, 1] = 360 - mirrored_points[:, 1]
        symmetric_points = np.row_stack((self.points, mirrored_points))

        return PolarDiagramPointcloud(symmetric_points)

    def add_points(self, new_pts, apparent_wind=False):
        """Adds additional points to the point cloud.

        Parameters
        ----------
        new_pts : array_like of shape (n, 3)
            New points to be added to the point cloud given as a sequence
            of points consisting of wind speed, wind angle and
            boat speed.

        apparent_wind : bool, optional
            Specifies if wind data is given in apparent wind.

            If `True`, data will be converted to true wind.

            Defaults to `False`.
        """
        if apparent_wind:
            new_pts = convert_apparent_wind_to_true(new_pts)
        else:
            new_pts = np.asarray_chkfinite(new_pts)
            if np.any((new_pts[:, 0] <= 0)):
                raise ValueError("`new_pts` has non-positive wind speeds")
            new_pts[:, 1] %= 360

        self._points = np.row_stack((self._points, new_pts))

    def _get_points(self, ws, range_):
        wa = []
        bsp = []
        cloud = self.points
        for w in ws:
            if not isinstance(w, tuple):
                w = (w - range_, w + range_)

            pts = cloud[
                np.logical_and(w[1] >= cloud[:, 0], cloud[:, 0] >= w[0])
            ][:, 1:]
            if not pts.size:
                raise RuntimeError(
                    f"no points with wind speed in range {w} found"
                )

            # sort for wind angles (needed for plotting methods)
            pts = pts[pts[:, 0].argsort()]

            wa.append(np.deg2rad(pts[:, 0]))
            bsp.append(pts[:, 1])

        if not wa:
            raise ValueError("there are no slices in the given range `ws`")

        return wa, bsp



"""
Interfaces and classes for data processing.

Subclasses can be used with the `PolarPipeline` class
of the `hrosailing.pipeline` module.
"""

from .datahandler import (
    ArrayHandler,
    CsvFileHandler,
    DataHandler,
    NMEAFileHandler,
)
from .filter import BoundFilter, Filter, QuantileFilter
from .imputator import FillLocalImputator, RemoveOnlyImputator
from .injector import ZeroInjector
from .interpolator import (
    ArithmeticMeanInterpolator,
    IDWInterpolator,
    ImprovedIDWInterpolator,
    Interpolator,
)
from .neighbourhood import (
    Ball,
    Cuboid,
    Ellipsoid,
    Neighbourhood,
    Polytope,
    ScalingBall,
)
from .regressor import LeastSquareRegressor, ODRegressor, Regressor
from .sampler import (
    ArchimedeanSampler,
    FibonacciSampler,
    Sampler,
    UniformRandomSampler,
)
from .smoother import AffineSmoother, LazySmoother, Smoother
from .weigher import (
    AllOneWeigher,
    CylindricMeanWeigher,
    CylindricMemberWeigher,
    FluctuationWeigher,
    FuzzyVariable,
    FuzzyWeigher,
    Weigher,
)

__all__ = [
    "ArrayHandler",
    "CsvFileHandler",
    "DataHandler",
    "NMEAFileHandler",
    "ArithmeticMeanInterpolator",
    "IDWInterpolator",
    "ImprovedIDWInterpolator",
    "Interpolator",
    "Ball",
    "Cuboid",
    "Ellipsoid",
    "Neighbourhood",
    "Polytope",
    "ScalingBall",
    "LeastSquareRegressor",
    "ODRegressor",
    "Regressor",
    "ArchimedeanSampler",
    "FibonacciSampler",
    "Sampler",
    "UniformRandomSampler",
    "AffineSmoother",
    "LazySmoother",
    "Smoother",
    "AllOneWeigher",
    "CylindricMeanWeigher",
    "CylindricMemberWeigher",
    "FluctuationWeigher",
    "FuzzyVariable",
    "FuzzyWeigher",
    "Weigher",
]



"""
Contains the Base class for the `Imputator` pipeline-component that can be used
to create custom imputators.
"""

from abc import ABC, abstractmethod
from datetime import timedelta

from hrosailing.core.statistics import ComponentWithStatistics


class Imputator(ComponentWithStatistics, ABC):
    """Base class for all imputator classes."""

    @abstractmethod
    def impute(self, data):
        """
        This method should be used, given data that possibly contains `None`
        values, to create data without any `None` values.

        Parameters
        ----------
        data : Data
            Data to be imputed.
        """

    def set_statistics(
        self, n_removed_cols, n_removed_rows, n_filled, data_dict
    ):
        super().set_statistics(
            n_removed_cols=n_removed_cols,
            n_removed_rows=n_removed_rows,
            n_filled_fields=n_filled,
            n_rows=data_dict.n_rows,
            n_cols=data_dict.n_cols,
        )


class RemoveOnlyImputator(Imputator):
    """
    Imputator that removes all entries containing None values.

    See also
    ----------
    `Imputator`
    """

    def impute(self, data):
        """
        Removes all entries containing None values.

        See also
        --------
        `Imputator.impute`
        """
        n_rows = data.n_rows
        remove_rows = [
            idx
            for idx, row in enumerate(data.rows(return_type=list))
            if None in row
        ]
        data.delete(remove_rows)

        self.set_statistics(0, n_rows - len(remove_rows), 0, data)

        return data


class FillLocalImputator(Imputator):
    """
    An `Imputator` which assumes that the data has been stored chronologically
    and contains the field `"datetime"`.
    Fills missing data by:
    - deleting columns that only contain `None` values,
    - deleting rows between two data-points that are far apart in time,
    - affine interpolation of datetime-stamps between two data-points which
        are not far apart in time,
    - filling data before and after a not-`None` value which are not too far apart
        in time according to certain functions (`fill_before` and `fill_after`),
    - filling data between a pair of not-`None` values which are not too far apart
        in time according to a certain function (`fill_between`).

    Parameters
    ----------
    fill_before : (str, object, float) -> object, optional
        The function which will be used to fill `None` fields before a not-`None`
        field.

        First argument is the attribute name, second is the value of the
        not-`None` field mentioned before, third argument is the relative
        position (in time) between the earliest applicable data and the
        mentioned not-`None` data. Returns the value that replaces `None`.

        Defaults to `lambda name, right, mu: right`.

    fill_between : (str, object, object, float) -> object
        The function which will be used to fill `None` fields between
        two not-`None` fields.

        First argument is the attribute `name`, second and third are the values
        of the not-`None` fields mentioned before in chronological order,
        last argument is the relative position (in time) between
        the two mentioned not-`None` data points.
        Returns the value to be filled.

        Defaults to `lambda name, left, right, mu: left`.

    fill_after : (str, object, float) -> object, optional
        The function which will be used to fill `None` fields after a not-`None`
        field.

        First argument is the attribute `name`, second is the value of the
        not-`None` field mentioned before, third argument is the relative
        position (in time) between the mentioned not-`None` data and the
        latest applicable data. Returns the value to be filled.

        Defaults to `lambda name, left, mu: left`.

    max_time_diff : datetime.timedelta, optional
        Two data points are treated as 'close in time' if their time difference
        is smaller than `max_time_diff`.

        Defaults to `2 minutes`.

    See also
    ----------
    `Imputator`
    """

    def __init__(
        self,
        fill_before=lambda name, right, mu: right,
        fill_between=lambda name, left, right, mu: left,
        fill_after=lambda name, left, mu: left,
        max_time_diff=timedelta(minutes=2),
    ):
        super().__init__()
        self._fill_before = fill_before
        self._fill_between = fill_between
        self._fill_after = fill_after
        self._max_time_diff = max_time_diff
        self._n_filled = 0

    def impute(self, data):
        """
        Creates a dictionary that does not contain `None` values by the procedure
        described above.

        Parameters
        ----------
        data : Data
            The `Data` object to be imputed. Has to support the key `datetime`.

        Returns
        -------
        data_dict : Data
            `data_dict` is the resulting `Data` object containing no `None` values.

        See also
        ----------
        `Imputator.impute`
        """
        self._n_filled = 0
        n_removed_cols = data.n_cols
        data.strip("cols")
        n_removed_cols -= data.n_cols

        data = self._interpolate_datetime(data)

        data, n_removed_rows = self._remove_rows(data)

        data = self._interpolate_other(data)

        # remove rows which still have None values
        remove_rows = [
            i
            for i, _ in enumerate(data["datetime"])
            if any(data[key][i] is None for key in data.keys())
        ]

        data.delete(remove_rows)
        n_removed_rows += len(remove_rows)

        self.set_statistics(
            n_removed_cols, n_removed_rows, self._n_filled, data
        )

        return data

    def _interpolate_datetime(self, data):
        last_dt, last_i = None, None
        for i, dt in enumerate(data["datetime"]):
            if dt is None:
                continue
            if last_dt is None:
                pass
            elif abs(dt - last_dt) > self._max_time_diff:
                continue
            else:
                # linear approximation of time in between
                for j in range(last_i + 1, i):
                    mu = (j - last_i) / (i - last_i)
                    data["datetime"][j] = last_dt + mu * (dt - last_dt)
            last_dt, last_i = dt, i

        return data

    def _remove_rows(self, data):
        remove_rows = [
            i for i, dt in enumerate(data["datetime"]) if dt is None
        ]
        data.delete(remove_rows)

        n_removed_rows = len(remove_rows)

        return data, n_removed_rows

    def _interpolate_other(self, data):
        # indices of not None values
        idx_dict = {
            key: [i for i, data in enumerate(data[key]) if data is not None]
            for key in data.keys()
        }

        datetime = data["datetime"]

        for key, idx in idx_dict.items():
            # fill every entry before the first not-None entry according to the
            # "fill before" function
            if key == "datetime":
                continue

            self._apply_fill_functions(datetime, idx, key, data)
            # try:
            #    start_idx = self._aplly_fill_before(datetime, idx, key, data)
            # except ValueError:
            #    continue

            # self._apply_fill_between(key, idx, data, datetime)

            # self._apply_fill_after(key, idx, data, datetime, start_idx)

        return data

    def _get_start_and_end_idx(self, datetime, i, idx):
        try:
            start_idx = max(
                j
                for j in idx
                if j < i and datetime[i] - datetime[j] < self._max_time_diff
            )
        except ValueError:
            start_idx = None
        try:
            end_idx = min(
                j
                for j in idx
                if j > i and datetime[j] - datetime[i] < self._max_time_diff
            )
        except ValueError:
            end_idx = None

        return start_idx, end_idx

    def _apply_fill_functions(self, datetime, idx, key, data):
        for i, entry in enumerate(data[key]):
            if i in idx:
                continue

            start_idx, end_idx = self._get_start_and_end_idx(datetime, i, idx)

            if start_idx is None and end_idx is None:
                continue

            try:
                range_too_big = (
                    datetime[end_idx] - datetime[start_idx]
                    > self._max_time_diff
                )
            except TypeError:
                range_too_big = False

            if start_idx is None:
                mu = (datetime[end_idx] - datetime[i]) / self._max_time_diff
                data[key][i] = self._fill_before(key, data[key][end_idx], mu)
                self._n_filled += 1
                continue

            if end_idx is None or range_too_big:
                mu = (datetime[i] - datetime[start_idx]) / self._max_time_diff
                data[key][i] = self._fill_after(key, data[key][start_idx], mu)
                self._n_filled += 1
                continue

            mu = (datetime[i] - datetime[start_idx]) / (
                datetime[end_idx] - datetime[start_idx]
            )
            data[key][i] = self._fill_between(
                key, data[key][start_idx], data[key][end_idx], mu
            )
            self._n_filled += 1



"""
Defines the base class `Injector` to be used as a component of the pipeline.

Also contains the predefined and ready to use injectors:

- `ZeroInjector`.
"""

from abc import ABC, abstractmethod

import numpy as np

from hrosailing.core.data import WeightedPoints
from hrosailing.core.statistics import ComponentWithStatistics


class Injector(ComponentWithStatistics, ABC):
    """
    Base class for all injector classes.
    """

    @abstractmethod
    def inject(self, weighted_points):
        """Method that should produce artificial weighted data points
        that are supposed to be appended to the original data.

        Parameters
        ----------
        weighted_points : WeightedPoints
            The original preprocessed points. `weighted_points.data` has to be an `ndarray`.

        Returns
        -------
        app_points : WeightedPoints
            Points to append to the original points.
        """

    def set_statistics(self, n_injected):
        super().set_statistics(n_injected=n_injected)


class ZeroInjector(Injector):
    """
    Injector which adds a fixed number of points with 0 boat speed.

    Parameters
    ----------
    n_zeros : int
        Number of artificial points to be added at 0 degree and at 360 degree
        respectively.

    See also
    ----------
    `Injector`
    """

    def __init__(self, n_zeros):
        super().__init__()
        self.n_zeros = n_zeros

    def inject(self, weighted_points):
        """Adds `n_zeros` points equally distributed in the `TWS` dimension with
        boat speed 0 and wind angle 0 and 360 respectively.


        Parameters
        ----------
        weighted_points : WeightedPoints
            The original preprocessed points with `weighted_points.data` an
            `numpy.ndarray` of shape (n, 3) containing at least one record.

        Returns
        -------
        app_points : WeightedPoints
            Points to append to the original points.

        See also
        ----------
        `Injector.inject`
        """

        data = weighted_points.data

        if not isinstance(data, np.ndarray):
            raise TypeError(
                "`weighted_points.data` should be an `numpy.ndarray` but is"
                f" {type(data)}."
            )
        if data.ndim != 2 or data.shape[1] != 3:
            raise ValueError(
                "`weighted_points.data` should be of shape (n, 3)"
            )
        if len(data.T) == 0:
            raise ValueError(
                "`weighted_points.data` should contain at least one record."
            )

        ws = data[:, 0]
        ws = np.linspace(min(ws), max(ws), self.n_zeros)

        zero = np.zeros(self.n_zeros)
        full = 360 * np.ones(self.n_zeros)
        zeros = np.column_stack((ws, zero, zero))
        fulls = np.column_stack((ws, full, zero))

        self.set_statistics(2 * self.n_zeros)
        return WeightedPoints(
            data=np.concatenate([zeros, fulls]),
            weights=np.ones(2 * self.n_zeros),
        )



"""
Functions for navigation and weather routing using polar diagrams.
"""


from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import List, Optional

import numpy as np
from scipy.integrate import solve_ivp, trapezoid
from scipy.spatial import ConvexHull

from hrosailing.models import InfluenceModel, WeatherModel


@dataclass
class Direction:
    """Dataclass to represent recommended sections of a sailing maneuver. Supports `str` method."""

    #: Right headed angle between the boat heading and the wind direction.
    #:   Same as TWA but from the boat's perspective.
    angle: float

    #: The recommended proportion of time needed to sail into this direction.
    #: Given as number between 0 and 1.
    proportion: float

    #: Type/Name of sail that should be hissed, when
    #: sailing in the direction (if existent)
    sail: Optional[str] = None

    def __str__(self):
        stc = (
            f"Sail with an angle of {self.angle} to the wind for "
            f"{self.proportion * 100} percent of the time"
        )

        if self.sail:
            stc += f", while hissing {self.sail}"

        return stc


def convex_direction(
    pd,
    ws,
    direction,
    im: Optional[InfluenceModel] = None,
    influence_data: Optional[dict] = None,
) -> List[Direction]:
    """Given a direction, computes the 'fastest' way to sail in
    that direction, assuming constant wind speed `ws`.

    If sailing straight into direction is the fastest way, function
    returns that direction. Otherwise, function returns two directions
    as well as their proportions, such that sailing into one direction for
    a corresponding proportion of a time segment and then into the other
    direction for a corresponding proportion of a time segment will be
    equal to sailing into `direction` but faster.

    Parameters
    ----------
    pd : PolarDiagram
        The polar diagram of the vessel.
    ws : int or float
        The current wind speed.
    direction : int or float
        Right handed angle between the heading of the boat and
        the negative of the wind direction.
        Numerically equals TWA, but interpreted from the perspective of the
        boat.
    im : InfluenceModel, optional
        The influence model used to consider additional influences
        on the boat speed.
        Defaults to `None`.
    influence_data : dict, optional
        Data containing information that might influence the boat speed
        of the vessel (e.g. current, wave height), to be passed to
        the used influence model.
        Will only be used if `im` is not `None`.
        Defaults to `None`.

    Returns
    -------
    edge : list of Directions
        Either just one `Direction` instance, if sailing into `direction`
        is the optimal way, or two `Direction` instances, that will be 'equal'
        to `direction`.

    Raises
    -------
    CruisingException
        If the given polar diagram slice can not be evaluated in the given
        direction. For example, this could be the case, if the polar diagram
        only has data for angles between 0 and 180 degrees.
    """
    labels, slices, info = pd.get_slices(ws, full_info=True)
    ws, wa, bsp = slices[0]
    if im:
        bsp = im.add_influence(pd, influence_data)
    bsp = np.array(bsp).ravel()
    wa = np.deg2rad(np.array(wa).ravel())

    polar_pts = np.column_stack(
        (bsp * np.cos(wa).ravel(), bsp * np.sin(wa).ravel())
    )
    conv = ConvexHull(polar_pts)
    vert = sorted(conv.vertices)

    wa = np.rad2deg(wa)

    for left, right in zip(vert, vert[1:]):
        if wa[left] <= direction <= wa[right]:
            i1, i2 = left, right
            edge = [Direction(wa[i1], 1), Direction(wa[i2], 1)]
            break
    else:
        i1, i2 = vert[0], vert[-1]
        if abs(wa[i1] - wa[i2]) < 180:
            raise ValueError(
                "The given direction is not supported by the given"
                " polar_diagram."
            )
        edge = [Direction(wa[i1], 1), Direction(wa[i2], 1)]

    if info:
        edge[0].sail = info[0][i1]
        edge[1].sail = info[0][i2]

    if edge[0] == direction:
        return [edge[0]]

    if edge[1] == direction:
        return [edge[1]]

    # direction lies on a common edge of polar diagram and convex hull
    if abs(i1 - i2) == 1 and edge[0].sail == edge[1].sail:
        return [edge[0]]

    lambda_ = (direction - wa[i2]) / (wa[i1] - wa[i2])
    if lambda_ > 1 or lambda_ < 0:
        lambda_ = (direction + 360 - wa[i2]) / (wa[i1] + 360 - wa[i2])

    edge[0].proportion = lambda_
    edge[1].proportion = 1 - lambda_
    return edge


def cruise(
    pd,
    start,
    end,
    wind,
    wind_fmt="ws_wan",
    im: Optional[InfluenceModel] = None,
    influence_data: Optional[dict] = None,
):
    """Calculates the fastest time and sailing direction for a vessel to reach `end`
    from `start`, under constant wind.

    If needed the function will calculate two directions as well as the
    time needed to sail in each direction to get to `end`.

    Wind has to be given by one of the following combinations of parameters:

    - `ws` and `wa_north`,
    - `ws`, `wa` and `hdt`,
    - `uv_grd`.

    Parameters
    ----------
    pd : PolarDiagram
        The polar diagram of the vessel.
    start : tuple of length 2
        Coordinates of the starting point of the cruising maneuver,
        given in longitude and latitude.
    end : tuple of length 2
        Coordinates of the end point of the cruising maneuver,
        given in longitude and latitude.
    wind: tuple
        Description of the wind. The exact interpretation depends on
        `wind_fmt`. See the description of `wind_fmt` for details.
    wind_fmt: {"ws_wan", ws_wa_hdt", "uv_grd"}, optional
        Specification how to interpret the parameter `wind`.

        - "ws_wan": `wind` is interpreted as
            `(true wind speed, wind angle relative to north)`,
        - "ws_wa_hdt": `wind` is interpreted as
            `(true wind speed, true wind angle, heading of the boat relative to north)`,
        - "uv_grd": `wind` is interpreted as `(u_grd, v_grd)` as can be read from
            a GRIB file.

        Defaults to "ws_wan".
    im : InfluenceModel, optional
        The influence model used to consider additional influences
        on the boat speed.
        Defaults to `None`.
    influence_data : dict, optional
        Further data to be passed to the used influence model.
        Only use data which does not depend on the wind and the heading.

        Will only be used if `im` is not `None`.

        Defaults to `None`.

    Returns
    -------
    directions : list of tuples
        Directions as well as the time (in hours) needed to sail along those,
        to get from start to end.

    Raises
    -------
    AttributeError
        If `wind_fmt` is not a supported string.
    """

    ws, wdir = _wind_relative_to_north(wind, wind_fmt)

    _, slices, _ = pd.get_slices(ws, full_info=True)
    ws, wa, bsp = slices[0]
    if im:
        for key, val in influence_data.items():
            influence_data[key] = [val] * len(wa)
        influence_data["TWS"] = [ws] * len(wa)
        influence_data["TWA"] = wa
        bsp = im.add_influence(pd, influence_data)
    bsp = np.array(bsp).ravel()

    rhc = _right_handing_course(start, end)

    heading = np.arccos(
        np.cos(rhc) * np.cos(wdir) + np.sin(rhc) * np.sin(wdir)
    )
    heading = 180 - np.rad2deg(heading)
    d1, *d2 = convex_direction(pd, ws, heading)

    dist = _great_earth_ellipsoid_distance(start, end)

    bsp1 = bsp[np.where(wa == d1.angle)[0]]
    if not d2:
        return [(d1.angle, float(dist / bsp1))]

    d2 = d2[0]
    bsp2 = bsp[np.where(wa == d2.angle)[0]]

    t = dist / (d1.proportion * bsp1 + d2.proportion * bsp2)
    t1, t2 = d1.proportion * t, d2.proportion * t

    return [(d1.angle, float(t1)), (d2.angle, float(t2))]


def cost_cruise(
    pd,
    start,
    end,
    start_time: datetime,
    wm: WeatherModel,
    cost_fun_dens=None,
    cost_fun_abs=lambda total_t, total_s: total_t,
    integration_method=trapezoid,
    im: Optional[InfluenceModel] = None,
    **ivp_kw,
):
    """Computes the total cost for traveling from a start position to an
    end position. To be precise:
    Let 'l' be the total distance of the start position and the end position,
    'cost' be a density cost function describing the costs generated at each
    point along the way (for example the indicator function for bad
    weather) and 'abs_cost' be a cost function describing the cost independent
    of the weather along the way.
    Note that 'abs_cost' only depends on the expected travel time and the
    expected travel distance.

    The method first approximates the travelled time (t)
    as a function dependent on distance travelled (s) by numerically solving
    the initial value problem

    .. math:: t(0) = 0, \\frac{dt}{ds} = \\frac{1}{bsp(s,t)}.

    Using this, it then uses numeric integration to predict the total costs as

    .. math::
        \\int_{0}^{l} cost(s, t(s)) \\,ds + abs\\\\_cost (t(l), l).

    Note that the costs in this mathematical description indirectly depend on
    weather forecast data, organized by a `WeatherModel`.
    Distances are computed using the mercator projection.

    Parameters
    ----------
    pd : PolarDiagram
        Polar diagram of the vessel.
    start : tuple of two floats
        Coordinates of the starting point.
    end : tuple of two floats
        Coordinates of the end point.
    start_time : datetime.datetime
        The time at which the traveling starts.
    wm : WeatherModel
        The weather model used. Needs to support the keys `"TWA"` and `"TWS"`.
    cost_fun_dens : callable, optional
        Function giving a cost density for given time as `datetime.datetime`,
        latitude as float, longitude as float and WeatherModel.
        `cost_fun_dens(t,lat,long,wm)` corresponds to `costs(s,t)` above.
        Defaults to `None`.
    cost_fun_abs : callable, optional
        Corresponds to `abs_costs`.
        Defaults to `lambda total_t, total_s: total_t`.
    integration_method : callable, optional
        Function that takes two (n,) arrays y, x and computes
        an approximate integral from that.
        Will only be used if `cost_fun_dens` is not `None`.
        Defaults to `scipy.integrate.trapezoid`.
    im : InfluenceModel, optional
        The influence model used to consider additional influences
        on the boat speed.
        Defaults to `None`.
    ivp_kw : Keyword arguments
        Keyword arguments which will be passed to `scipy.integrate.solve_ivp`
        in order to solve the initial value problem described above.

    Returns
    -------
    cost : float
        The total cost calculated as described above.
    """
    # pylint: disable=too-many-locals

    lat_mp = (start[0] + end[0]) / 2
    proj_start = _mercator_proj(start, lat_mp)
    proj_end = _mercator_proj(end, lat_mp)
    total_s = np.linalg.norm(proj_end - proj_start)

    hdt = _right_handing_course(start, end)

    # define derivative of t by s
    def dt_ds(s, t):
        pos = proj_start + s / total_s * (proj_end - proj_start)
        inv_bsp = _get_inverse_bsp(
            pd, pos, hdt, t[0], lat_mp, start_time, wm, im
        )
        return inv_bsp

    t_s = solve_ivp(
        fun=dt_ds,
        t_span=(0, np.linalg.norm(proj_start - proj_end)),
        y0=np.zeros(1),
        **ivp_kw,
    )

    # calculate absolute cost and return it if sufficient
    total_t = t_s.y[0][-1]  # last entry of IVP solution
    absolute_cost = cost_fun_abs(total_t, total_s)

    if not cost_fun_dens:
        return absolute_cost

    # calculate the integral described in the doc string
    pos_list = [
        proj_start + s / total_s * (proj_end - proj_start) for s in t_s.t
    ]
    lat_long_list = [_inverse_mercator_proj(pos, lat_mp) for pos in pos_list]
    t_list = [start_time + timedelta(hours=t) for t in t_s.y[0]]

    costs = [
        cost_fun_dens(t, lat, long, wm)
        for t, (lat, long) in zip(t_list, lat_long_list)
    ]

    return absolute_cost + integration_method(costs, t_s.t)


def isochrone(
    pd,
    start,
    start_time,
    direction,
    wm: WeatherModel,
    total_time=1,
    min_nodes=100,
    im: Optional[InfluenceModel] = None,
):
    """
    Estimates the maximum distance that can be reached from a given start
    point in a given amount of time without tacks and jibes.
    This is done by sampling the position space and using mercator projection.
    Weather forecast data, organized by a `WeatherModel` and an `InfluenceModel`,
    is included in the computation.

    Parameters
    ----------
    pd : PolarDiagram
        The polar diagram of the used vessel.
    start : 2-tuple of floats
        The latitude and longitude of the starting point.
    start_time : datetime.datetime
        The time at which the traveling starts.
    direction : float
        The angle between North and the direction in which we aim to travel.
    wm : WeatherModel
        The weather model used. Needs to support the keys 'TWA' and 'TWS'.
    total_time : float, optional
        The time in hours that the vessel is supposed to travel
        in the given direction.
        Defaults to `1`.
    min_nodes : int, optional
        The minimum amount of sample points to sample the position space.
        Defaults to `100`.
    im : InfluenceModel, optional
        The influence model used.

        Defaults to `None`.

    Returns
    -------
    end : 2-tuple of floats
        Latitude and longitude of the position that is reached when traveling
        `total_time` hours in the given direction.

    s : float
        The length of the way traveled from start to end.
    """
    # estimate first sample points as equidistant points

    lat_mp = start[0]
    proj_start = _mercator_proj(start, lat_mp)
    arc = np.pi * (1 / 2 - direction / 180)
    v_direction = np.array([np.cos(arc), np.sin(arc)])

    def dt_ds(s, t):
        pos = proj_start + s * v_direction
        return _get_inverse_bsp(
            pd, pos, direction, t, lat_mp, start_time, wm, im
        )

    # supposed boat speed for first estimation is 5 knots
    step_size = 5 * total_time / min_nodes
    s, t, steps = 0, 0, 0

    der = 0  # debug

    while t < total_time or steps < min_nodes:
        if t >= total_time:
            # start process again with smaller step size
            step_size *= steps / min_nodes
            s, t, steps = 0, 0, 0
            continue
        der = dt_ds(s, t)
        s += step_size
        t += der * step_size
        steps += 1

    # we end up with s, t such that t >= total_time and steps > min_nodes
    # still need to correct the last step such that t == total_time

    s = (total_time + der * s - t) / der

    proj_end = proj_start + s * v_direction
    end = _inverse_mercator_proj(proj_end, lat_mp)

    return end, s


def _inverse_mercator_proj(pt, lat_mp):
    """
    Computes point from its mercator projection with reference point `lat_mp`.
    """
    x, y = pt / 69
    return x + lat_mp, 180 / np.pi * np.arcsin(np.tanh(y))


def _mercator_proj(pt, lat_mp):
    """
    Computes the mercator projection with reference point `lat_mp` of a point.
    """
    lat, long = pt

    # 69 nautical miles between two latitudes
    return 69 * np.array(
        [(lat - lat_mp), np.arcsinh(np.tan(np.pi * long / 180))]
    )


def _get_inverse_bsp(pd, pos, hdt, t, lat_mp, start_time, wm, im):
    """"""
    lat, long = _inverse_mercator_proj(pos, lat_mp)
    time = start_time + timedelta(hours=t)
    try:
        data = wm.get_weather((time, lat, long))
    except ValueError:
        return 1000  # maybe a better default?
    data["HDT"] = hdt
    if im:
        data = {key: [val] for key, val in data.items()}
        bsp = im.add_influence(pd, data)[0]
    else:
        ugrid, vgrid = data["UGRID"], data["VGRID"]
        tws, twa = _uvgrid_to_tw(ugrid, vgrid, hdt)
        bsp = pd(tws, twa)

    if bsp != 0:
        return 1 / bsp

    return 0


def _right_handing_course(a, b):
    """Calculates course between two points on the surface of the earth
    relative to true north.
    """
    numerator = np.cos(a[1]) * np.sin(b[1]) - np.cos(a[0] - b[0]) * np.cos(
        b[1]
    ) * np.sin(a[1])
    denominator = np.cos(a[0] - b[0]) * np.cos(a[1]) * np.cos(b[1]) + np.sin(
        a[1]
    ) * np.cos(b[1])

    return np.arccos(numerator / np.sqrt(1 - denominator**2))


def _wind_relative_to_north(wind, wind_fmt):
    """Calculates the wind speed and the wind direction relative to true north.

    Parameters
    ----------
    wind: tuple
        Description of the wind. The exact interpretation depends on
        `wind_fmt`. See the description of `wind_fmt` for details.

    wind_fmt: {"ws_wan", "ws_wa_hdt", "uv_grd"}
        Specification how to interpret the parameter `wind`.

        - "ws_wan": `wind` is interpreted as
            (true wind speed, wind angle relative to north),
        - "ws_wa_hdt": `wind` is interpreted as
            (true wind speed, true wind angle,
            heading of the boat relative to north),
        - "uv_grd": `wind` is interpreted as (u_grd, v_grd) as can be read from
            a GRIB file.

    Returns
    -------
    ws : float,
        The current wind speed.

    ndir : float between 0 and 360

    Raises
    --------
    AttributeError
        If `wind_fmt` is not a supported string.
    """
    if wind_fmt == "ws_wan":
        return wind

    if wind_fmt == "ws_wa_hdt":
        ws, wa, hdt = wind
        return ws, (hdt - wa) % 360

    if wind_fmt == "uv_grd":
        wind = np.array(wind)
        u, v = wind
        return np.linalg.norm(wind), 180 / np.pi * np.arctan2(v, u)

    raise AttributeError(f"wind_fmt '{wind_fmt}' is not supported")

    # grib data:
    # wdir = 180 / np.pi * np.arctan2(vgrd, ugrd) + 180

    # twa + bd:
    # wdir = (rwSK + twa) % 360 ?


def _uvgrid_to_tw(ugrid, vgrid, hdt):
    """Calculates the true wind speed and wind angle from given grib data."""
    tws = np.sqrt(ugrid**2 + vgrid**2)
    wa = (180 + 180 / np.pi * np.arctan2(vgrid, ugrid)) % 360
    twa = (hdt - wa) % 360
    return tws, twa


EARTH_FLATTENING = 1 / 298.257223563
EQUATOR_CIRCUMFERENCE = 40075.017


def _great_earth_ellipsoid_distance(a, b):
    """Calculates the distance on the surface for two points on the
    earth surface.
    """
    f = (a[1] + b[1]) / 2
    g = (a[1] - b[1]) / 2
    lat = (a[0] - b[0]) / 2

    s = (np.sin(g) * np.cos(lat)) ** 2 + (np.cos(f) * np.sin(lat)) ** 2
    c = (np.cos(g) * np.cos(lat)) ** 2 + (np.sin(f) * np.sin(lat)) ** 2

    omega = np.deg2rad(np.arctan(np.sqrt(s / c)))
    d = EQUATOR_CIRCUMFERENCE * omega / np.pi

    t = np.sqrt(s * c) / omega
    h_1 = (3 * t - 1) / (2 * c)
    h_2 = (3 * t + 1) / (2 * s)

    dist = d * (
        1
        + EARTH_FLATTENING
        * (
            h_1 * (np.sin(f) * np.cos(g)) ** 2
            - h_2 * (np.cos(f) * np.sin(g)) ** 2
        )
    )
    return dist



"""
Functions for plotting support of instances of `hrosailing` objects.
Currently, plots of `PolarDiagram` instances are supported.
"""

import hrosailing.plotting.projections
from hrosailing.plotting.projections import (
    plot_3d,
    plot_color_gradient,
    plot_flat,
    plot_polar,
    scatter_3d,
    scatter_flat,
    scatter_polar,
)

__all__ = [
    "plot_polar",
    "plot_flat",
    "plot_3d",
    "plot_color_gradient",
    "scatter_flat",
    "scatter_polar",
    "scatter_3d",
]



"""
Functions for navigation and weather routing using polar diagrams.
"""

from .basics import Direction, convex_direction, cost_cruise, cruise, isochrone

__all__ = [
    "Direction",
    "convex_direction",
    "cruise",
    "cost_cruise",
    "isochrone",
]



"""
Functions for the handling of statistics.
"""


class ComponentWithStatistics:
    """
    Interface class for pipelinecomponents enabling to save and handle run statistics.
    """

    def __init__(self):
        self._statistics = {}

    def set_statistics(self, **kwargs):
        """
        Sets the statistics dict corresponding to the keyword arguments.
        Supposed to be overwritten by inheriting classes with stronger requirements.
        """
        self._statistics = kwargs

    def get_latest_statistics(self):
        """
        Return
        ------
        statistics: dict
            The statistics of the latest run of the handler.
        """
        return self._statistics



"""
Contains the baseclass `Expander` used in the `PolarPipeline` class,
that can also be used to create custom expanders.

Also contains predefined and usable expanders:

- `LazyExpander`,
- `WeatherExpander`.
"""

from abc import ABC, abstractmethod

from hrosailing.core.data import Data
from hrosailing.core.statistics import ComponentWithStatistics


class Expander(ComponentWithStatistics, ABC):
    """Base class for all expander classes."""

    @abstractmethod
    def expand(self, data):
        """
        Should be used to expand the data by more data fields resulting from additional data sources.

        Parameters
        ----------
        data : Data
            The data that will be expanded.

        Returns
        -------
        data : Data
            The processed data.
        """

    def set_statistics(self, data):
        super().set_statistics(n_rows=data.n_rows, n_cols=data.n_cols)


class LazyExpander(Expander):
    """
    Expander that doesn't do anything.

    See also
    ----------
    `Expander`
    """

    def expand(self, data):
        """
        See also
        --------
        `Expander.expand`
        """
        self.set_statistics(data)
        return data


class WeatherExpander(Expander):
    """
    Expander that uses a weather model to add weather data to given data
    if the fields `datetime`, `lat` and `lon` are defined.

    Parameters
    ----------
    weather_model : WeatherModel
        A suitable weather model (yielding weather information for the required times, latitudes and longitudes).

    exception_handling_mode: {"delete", "ignore"}, optional
        Describes how to handle weather cases which throw a ValueError.

        - `"delete"` : delete the occurrences,
        - `"ignore"` : fill the corresponding records with `None`.

        Defaults to `"delete"`.

    See also
    ----------
    `Expander`
    """

    def __init__(self, weather_model, exception_handling_mode="delete"):
        super().__init__()
        self._weather_model = weather_model
        self._exception_handling_mode = exception_handling_mode

    def expand(self, data):
        """
        Expands given data by the procedure described above.

        See also
        --------
        `Expander.expand`
        """
        weather_data = self._get_weather(data)
        weather_data.hrosailing_standard_format()
        weather_data = self._mark_duplicate_keys(weather_data, data)

        data.update(weather_data)
        data.hrosailing_standard_format()
        self.set_statistics(data)
        return data

    def _get_weather(self, data):
        weather_keys = None
        none_idxs = []
        weather_list = []
        for idx, (datetime, lat, lon) in enumerate(
            data.rows(["datetime", "lat", "lon"], return_type=tuple)
        ):
            try:
                weather = self._weather_model.get_weather([datetime, lat, lon])
                weather_list.append(weather)
                if weather_keys is None:
                    weather_keys = weather.keys()
            except ValueError:
                if self._exception_handling_mode == "ignore":
                    weather_list.append(None)
                none_idxs.append(idx)

        if self._exception_handling_mode == "delete":
            data.delete(none_idxs)
        else:
            for idx in none_idxs:
                weather_list[idx] = {key: None for key in weather_keys}

        return Data.concatenate(weather_list)

    def _mark_duplicate_keys(self, weather_data, data):
        for key in list(weather_data.keys()):
            if key in data.keys():
                weather_data.rename(key, f"{key}(duplicate)")
        return weather_data



# pylint: disable=missing-module-docstring

import csv

import numpy as np

from ._basepolardiagram import PolarDiagram
from ._polardiagramtable import PolarDiagramTable
from ._reading import from_csv


class PolarDiagramMultiSails(PolarDiagram):
    """A class to represent, visualize and work with
    a polar diagram made up of multiple sets of sails,
    represented by a `PolarDiagramTable`.


    Class methods aren't fully developed yet. Take care when
    using this class.

    Parameters
    ----------
    pds : Sequence of PolarDiagramTable objects
        Polar diagrams belonging to different sets of sails,
        given as tables, that share the same wind speeds.

    sails : Sequence, optional
        Custom names for the sails. Length should be equal to `pds`.
        If it is not equal it will either be cut off at the appropriate
        length or will be addended with `"Sail i"` to the appropriate length.

        Only important for the legend of plots or the `to_csv()`-method.

        If nothing is passed, the names will be `"Sail i"`, i = 0...n-1,
        where `len(pds) = n`.

    Attributes
    ---------
    sails (property) : list of str
        A read only version of the list of sails.

    diagrams (property) : list of `PolarDiagram`
        The different polar diagrams.

    SeeAlso
    -------
    `PolarDiagram`
    """

    def __init__(self, pds, sails=None):
        if sails is None:
            sails = [f"Sail {i}" for i in range(len(pds))]
        elif len(sails) < len(pds):
            sails = list(sails) + [
                f"Sail {i}" for i in range(len(sails), len(pds))
            ]
        elif len(sails) > len(pds):
            sails = list(sails)
            sails = sails[: len(pds)]

        self._sails = list(sails)
        self._diagrams = list(pds)

    @property
    def sails(self):
        return self._sails

    @property
    def diagrams(self):
        return self._diagrams

    def __getitem__(self, item) -> PolarDiagramTable:
        """"""
        try:
            index = self.sails.index(item)
        except ValueError as ve:
            raise ValueError("`item` is not a name of a sail") from ve

        return self._diagrams[index]

    def __str__(self):
        tables = [str(pd) for pd in self._diagrams]
        names = [str(sail) for sail in self._sails]
        out = []
        for name, table in zip(names, tables):
            out.append(name)
            out.append("\n")
            out.append(table)
            out.append("\n\n")

        return "".join(out)

    def __repr__(self):
        return (
            f"PolarDiagramMultiSails({[repr(diagram) for diagram in self._diagrams]},"
            f" {self.sails})"
        )

    def __call__(self, ws, wa):
        return max(diagram(ws, wa) for diagram in self._diagrams)

    @property
    def default_points(self):
        return np.row_stack([pd.default_points for pd in self._diagrams])

    def get_slices(self, ws=None, n_steps=None, full_info=False, **kwargs):
        """
        Parameters
        ---------
        **kwargs :
            The keyword arguments are forwarded to the `ws_to_slices` methods
            of the polardiagrams in `self.diagrams`.

        Returns
        --------
        slices : list of slices
            Concatenations of the slices of the polardiagrams in
            `self.diagrams`.

        See also
        -------
        `PolarDiagram.get_slices`
        """
        # make codefactor happy by doing something, feel free to refactor.
        kwargs["a"] = 5
        del kwargs["a"]
        return super().get_slices(ws, n_steps, full_info, **kwargs)

    def ws_to_slices(self, ws, **kwargs):
        """
        See also
        -------
        `PolarDiagramMultisails.get_slices`
        `PolarDiagram.ws_to_slices`
        """
        all_slices = [pd.ws_to_slices(ws, **kwargs) for pd in self._diagrams]
        slices = [
            np.column_stack(slice_collection)
            for slice_collection in zip(*all_slices)
        ]
        return slices

    def get_slice_info(self, ws, slices, **kwargs):
        """
        Returns
        --------
        info : list of lists
            Contains the sail name used for each record. Organized in the
            same manner as the slices.

        See also
        ----------
        `PolarDiagram.get_slices`
        """
        all_slices = [pd.ws_to_slices(ws, **kwargs) for pd in self._diagrams]
        slice_infos = [
            [
                sail
                for sail, slice in zip(self.sails, slice_collection)
                for _ in range(slice.shape[1])
            ]
            for slice_collection in zip(*all_slices)
        ]
        return slice_infos

    def to_csv(self, csv_path):
        """Creates a .csv file with delimiter ',' and the
        following format:

            `PolarDiagramMultiSails
            Sails:.....`

            The sub-diagrams are automatically saved in the files
            `{csv_path}_sail_{i}.csv`

        Parameters
        ----------
        csv_path : path_like
            Path to a .csv file or where a new .csv file will be created.
        """
        with open(csv_path, "w", newline="", encoding="utf-8") as file:
            csv_writer = csv.writer(file, delimiter=":")
            csv_writer.writerow([self.__class__.__name__])
            csv_writer.writerow(["Sails"] + self.sails)

        for i, diagram in enumerate(self._diagrams):
            sub_csv_path = self._get_sub_csv_path(csv_path, i)
            diagram.to_csv(sub_csv_path)

    @staticmethod
    def _get_sub_csv_path(path, i):
        path = "_".join([path.split(".")[0], "sail", str(i)])
        return f"{path}.csv"

    @classmethod
    def __from_csv__(cls, file):
        path = file.name
        csv_reader = csv.reader(file, delimiter=":")
        sails = next(csv_reader)[1:]
        diagrams = []
        for i, sail in enumerate(sails):
            sub_path = PolarDiagramMultiSails._get_sub_csv_path(path, i)
            diagrams.append(from_csv(sub_path))
        return PolarDiagramMultiSails(diagrams, sails)

    def symmetrize(self):
        """Constructs a symmetric version of the polar diagram, by
        mirroring each `PolarDiagramTable` at the 0° - 180° axis and
        returning a new instance. See also the `symmetrize()`-method
        of the `PolarDiagramTable` class.

        Warning
        -------
        Should only be used if all the wind angles of the `PolarDiagramTables`
        are each on one side of the 0° - 180° axis, otherwise this can lead
        to duplicate data, which can overwrite or live alongside old data.
        """
        pds = [pd.symmetrize() for pd in self.diagrams]
        return PolarDiagramMultiSails(pds, self._sails)

    @property
    def default_slices(self):
        all_defaults = np.concatenate(
            [pd.default_slices for pd in self._diagrams]
        )
        return np.linspace(all_defaults.min(), all_defaults.max(), 20)



"""
Defines the `InfluenceModel` abstract base class that can be used
to create custom influence models.

Subclasses of `InfluenceModel` can be used with

- the `PolarPipeline` class in the `hrosailing.pipeline` module,
- various functions in the `hrosailing.cruising` module.
"""

from abc import ABC, abstractmethod

import numpy as np

from hrosailing.core.computing import (
    convert_apparent_wind_to_true,
    data_dict_to_numpy,
)
from hrosailing.core.statistics import ComponentWithStatistics


class InfluenceModel(ComponentWithStatistics, ABC):
    """Base class for all influence model classes."""

    @abstractmethod
    def remove_influence(self, data):
        """This method should be used to create a `numpy.ndarray`
        output from a given `Data` object where the columns correspond to wind speed, wind angle and
        boat speed respectively.

        Parameters
        ----------
        data : Data
            Has to contain at least keys for wind speed, wind angle
            and either speed over ground, speed over water or boat speed.

        Returns
        -------
        out : numpy.ndarray
        """

    @abstractmethod
    def add_influence(self, pd, influence_data):
        """This method should be used, given a polar diagram and a
        dictionary, to obtain a modified boat speed of that given
        in the polar diagram, based on the influences presented in
        the given dictionary, such as wave height, underlying currents etc.

        Parameters
        ----------
        pd : PolarDiagram

        influence_data : Data or dict or list of Data or list of dict
            Further influences to be considered.

        Returns
        -------
        speeds : float or list of floats
            The boat speed if `influence_data` contained values,
            a list of respective boat speeds if `influence_data` contained
            lists.
        """

    def fit(self, training_data):
        """
        This method can be overwritten in order
        to fit parameters of the influence
        model to given training data.

        Parameters
        ----------
        training_data : Data
        """


class IdentityInfluenceModel(InfluenceModel):
    """An influence model which ignores most influences and just calculates
    the true wind if only the apparent wind is given.
    IF `"BSP"` is not provided by the data, `"SOG"` is used instead.

    See also
    ----------
    `InfluenceModel`
    """

    def remove_influence(self, data: dict):
        """
        Ignores most influences as described above.

        Parameters
        ----------
        data : dict
            Data dictionary, must either provide `"BSP"` or `"SOG"` key as well as
            either the keys `"TWS"`, `"TWA"` or `"AWS"`, `"AWA"`.

        Returns
        -------
        (n,3)-array

        See also
        --------
        `InfluenceModel.remove_influence`
        """
        return _get_true_wind_data(data)

    def add_influence(self, pd, influence_data: dict):
        """
        Ignores most influences and strictly calculates the boat speed using the
        polar diagram.

        Parameters
        ----------
        pd : PolarDiagram

        influence_data : dict
            Either a dictionary of lists or a dictionary of values containing
            one or more sets of influence data.
            At least the keys `"TWS"` and `"TWA"` have to be provided.

        See also
        --------
        `InfluenceModel.add_influence`
        """
        if isinstance(influence_data["TWS"], list):
            wind = zip(influence_data["TWS"], influence_data["TWA"])
            speed = [pd(ws, wa) for ws, wa in wind]
        else:
            ws, wa = influence_data["TWS"], influence_data["TWA"]
            speed = pd(ws, wa)
        return speed

    def fit(self, training_data: dict):
        """
        Does nothing.

        See also
        --------
        `InfluenceModel.fit`
        """


class WindAngleCorrectingInfluenceModel(InfluenceModel):
    """
    An influence model which corrects a structural measurement error in `"TWA"`.

    Parameters
    ----------
    wa_shift: int or float, optional
        Difference between real wind angle and measured wind angle (correction value).

        Defaults to `0`.

    interval_size: int or float, optional
        Interval size used in the fitting method.

        Defaults to `30`.

    See also
    ----------
    `InfluenceModel`
    """

    def __init__(self, interval_size=30, wa_shift=0):
        super().__init__()
        self._wa_shift = wa_shift
        self._interval_size = interval_size

    def remove_influence(self, data):
        """
        Removes the correction value from the data.

        See also
        -------
        `InfluenceModel.remove_influence`
        """
        wind_data = _get_true_wind_data(data)
        wa = wind_data[:, 1]
        wind_data[:, 1] = (wa - self._wa_shift) % 360
        return wind_data

    def add_influence(self, pd, influence_data):
        """
        Adds the correction value to the data.

        See also
        -------
        `InfluenceModel.add_influence`

        """
        if isinstance(influence_data["TWS"], (list, np.ndarray)):
            wind = zip(influence_data["TWS"], influence_data["TWA"])
            speed = [pd(ws, (wa + self._wa_shift) % 360) for ws, wa in wind]
        else:
            ws, wa = influence_data["TWS"], influence_data["TWA"]
            speed = pd(ws, wa)
        return speed

    def fit(self, training_data):
        """
        The wind angle with the lowest density of measured wind angles is assumed to be the
        actual zero. The data density is computed using gauss kernel functions :math:`e^{\\frac{wa - wa'}{l}}`,
        where :math:`l` can be seen as the size of an interval on which the gauss kernel is stretched.

        Parameters
        ----------
        training_data: Data containing key `"TWA"`

        See also
        --------
        `InfluenceModel.fit`
        """
        wind_angles = training_data["TWA"]
        wind_speeds = training_data["TWS"]
        sample = np.linspace(0, 359, 10 * 360)
        counts = [
            sum(
                other_ws
                * np.exp(
                    -(((abs(wa - other_wa) % 360) / self._interval_size) ** 2)
                )
                for other_wa, other_ws in zip(wind_angles, wind_speeds)
            )
            for wa in sample
        ]
        min_angle, _ = min(zip(sample, counts), key=lambda x: x[1])
        self._wa_shift = min_angle
        self.set_statistics(wa_shift=min_angle)


def _get_true_wind_data(data: dict):
    speed = "BSP" if "BSP" in data else "SOG"
    if "TWA" in data and "TWS" in data:
        if isinstance(data["TWS"], list):
            return data_dict_to_numpy(data, ["TWS", "TWA", speed])

        return np.array([data["TWS"], data["TWA"], data[speed]])

    if "AWA" in data and "AWS" in data:
        apparent_data = data_dict_to_numpy(data, ["AWS", "AWA", speed])
        return convert_apparent_wind_to_true(apparent_data)

    raise ValueError(
        "No sufficient wind data is given in order to apply influence"
        "model. Either give 'AWA' and 'AWS' or 'TWA' and 'TWS'"
    )



"""
Classes, methods and interfaces relevant for all other modules.
"""

import hrosailing.core.computing
import hrosailing.core.constants
import hrosailing.core.data
import hrosailing.core.exceptions
import hrosailing.core.modelfunctions
import hrosailing.core.statistics

__all__ = [
    "modelfunctions",
    "statistics",
    "computing",
    "data",
    "constants",
    "exceptions",
]



# pylint: disable=missing-module-docstring

from abc import ABC, abstractmethod

import numpy as np


class PolarDiagram(ABC):
    """Base class for all polar diagrams.

    Attributes
    -----------
    default_slices (abstract property) : numpy.ndarray of shape (n)
        Should give the default windspeeds for `get_slices`.

    default_points (abstract property) : numnpy.ndarray of shape (n, 3)
        Should give the 'points' (consisting of wind speed, wind angle and boat speed) at which the
        polar diagram is evaluated by default.
    """

    @abstractmethod
    def to_csv(self, csv_path):
        """This method should, given a path, write a .csv file in
        the location, containing human-readable information about the
        polar diagram object that called the method.

        Parameters
        ----------
        csv_path: str
            Path of the created csv file.
        """

    @classmethod
    @abstractmethod
    def __from_csv__(cls, file):
        """"""

    @abstractmethod
    def __call__(self, ws, wa):
        pass

    @abstractmethod
    def symmetrize(self):
        """This method should return a new `PolarDiagram` object that is a
        symmetric (i.e. mirrored along the 0 - 180° axis) version of the
        `PolarDiagram` object called the method.
        """

    def get_slices(self, ws=None, n_steps=None, full_info=False, **kwargs):
        """This method should produce a list of 'slices' describing the
        performance of specified wind_speed as well as wind angles and
        corresponding boat speeds, that reflect how the vessel behaves near
        the given wind speeds.

        Parameters
        ----------
        ws : int, float or iterable thereof, optional
            The wind speeds corresponding to the requested slices.

            Defaults to `self.default_slices`

        n_steps : int, optional
            If set, a total of `n_steps` wind_speeds between each value given
            in `ws` is taken into account. For example `n_steps = 1` adds all
            midpoints.

            Defaults to 0.

        full_info : bool, optional
            Specifies wether the additional value `info` will be
            returned or not

            Defaults to `False`.

        **kwargs :
            Additional keyword arguments whose functionality depends on the
            inheriting class. Are forwarded to `ws_to_slices`.

        Returns
        ---------
        labels : np.ndarray
            The wind speeds corresponding to the slices

        slices : list of `numpy.ndarray` of shape (3, *)
            A list of slices.
            A slice stores row wise data points consisting of actual
            wind speed, wind angle and boat speed.
            Note that the actual wind speed can differ (slightly) from the
            wind speed corresponding to the slice, depending on the inheriting
            class.
        """
        ws = self._get_windspeeds(ws, n_steps)
        kwargs["full_info"] = full_info
        slices = self.ws_to_slices(ws, **kwargs)
        if full_info:
            info = self.get_slice_info(ws, slices, **kwargs)
            return ws, slices, info
        return ws, slices

    def _get_windspeeds(self, ws, n_steps):
        if ws is None:
            ws = self.default_slices
        if isinstance(ws, (int, float)):
            return [ws]
        try:
            all_numbers = all(
                isinstance(ws_, (int, float, np.integer, np.floating))
                for ws_ in ws
            )
        except TypeError as exp:
            raise TypeError(
                "`ws` has to be an int a float or an iterable"
            ) from exp
        if not all_numbers:
            raise TypeError(
                "If `ws` is an iterable, it needs to iterate over int or float"
            )
        if n_steps is None:
            return np.array(ws)
        if n_steps <= 0:
            raise ValueError("`n_steps` has to be positive")
        return np.concatenate(
            [
                np.linspace(ws_before, ws_next, n_steps + 2)[:-1]
                for ws_before, ws_next in zip(ws, ws[1:])
            ]
            + [[ws[-1]]]
        )

    def _get_wind(self, wind):
        if isinstance(wind, np.ndarray):
            if wind.shape[1] == 2:
                return wind
            if wind.shape[0] == 2:
                return wind.T
            raise ValueError(
                "`wind` should be a tuple or an array with a dimension of"
                f" shape 2,\ngot an array of shape {wind.shape} instead."
            )
        if isinstance(wind, tuple):
            if len(wind) != 2:
                raise ValueError(
                    "`wind` should be a tuple of size 2 or an array,\n"
                    f"got a tuple of size {len(wind)} instead."
                )
            ws, wa = wind
            ws, wa = np.meshgrid(ws, wa)
            return np.array(list(zip(ws.ravel(), wa.ravel())))
        raise TypeError(
            f"`wind` should be a tuple or an array, got {type(wind)} instead."
        )

    @property
    @abstractmethod
    def default_points(self):
        pass

    @property
    @abstractmethod
    def default_slices(self):
        pass

    @abstractmethod
    def ws_to_slices(self, ws, **kwargs):
        """
        Should produce slices for the windspeeds given in ws.

        Parameters
        ---------
        ws : 1 dimensional numpy.ndarray
            The wind speeds corresponding to the requested slices.

        **kwargs :
            Further keyword arguments that may be used by custom
            implementations. Are forwarded by the `get_slices` method.
            Supports the key 'full_info'.

        Returns
        --------
        slices : list of (3, *) numpy.ndarrays
            List of the requested slices. The three rows of a slice
            should correspond to the actual
            wind speeds, the wind angles and the boat speeds

        See also
        -----------
        `get_slices`
        """

    def get_points(self, wind=None):
        """
        Returns a read only version of all relevant points specified by wind

        Parameters
        ----------
        wind : (2, d) or (d, 2) numpy.ndarray or tuple of 2 (1, d) numpy.ndarray or `None`, optional
            A specification of the wind.

            - If given as a (2, d) or (d, 2) numpy.ndarray the columns (rows)
                are interpreted as different wind records
                (for d=2 we assume the shape (d, 2))
            - If given as a tuple, the first array defines the wind speed
             resolution, the second array defines the wind angle resolution.
             The wind records are all combinations thereof.

        Returns
        ---------
        points : (m, 3) numpy.ndarray
            Row-wise contains all records specified in `wind`
            consisting of wind speed, wind angle, boat speed.
            If `wind` is not specified, `self.default_points` is returned.
        """

        if wind is None:
            return self.default_points
        wind = self._get_wind(wind)
        bsps = [self(wa, ws) for wa, ws in wind]
        return np.column_stack([wind, bsps])

    def get_slice_info(self, ws, slices, **kwargs):
        """
        Should produce additional information about slices depending on the
        inheriting class which is returned when the `full_info` parameter of
        the `get_slices` method is set to `True`. If no further information
        is known, return None.

        Parameters
        ----------
        ws : 1 dimensional numpy.ndarray
            The wind speeds corresponding to the requested slices.

        slices : list of (3, *) `np.ndarray`
            Slices produced by `ws_to_slices`

        **kwargs :
            Further keyword arguments that may be used by custom
            implementations. Are forwarded by the `get_slices` method.
            Supports the key 'full_info'.

        Returns
        -------
        info : list of lists
            List of additional information for each record of each slice
            organized in the same manner as a slice.

        See also
        -----------
        `get_slices`
        """
        # I do something with the variables to make codefactor happy, feel free to refactor
        kwargs["ws"] = ws
        kwargs["slices"] = slices
        return None



"""
Contains a base class and several concrete implementations of globe models
which can be used to perform calculations on a three-dimensional model of the
globe using latitude/longitude coordinates.
"""

from abc import ABC, abstractmethod

import numpy as np


class GlobeModel(ABC):
    """
    Abstract base class of globe models.
    Contains methods to translate between points given in latitude and
    longitude coordinates and points on a map or a three-dimensional globe.
    """

    @abstractmethod
    def project(self, points):
        """
        Calculates the three-dimensional points corresponding to a given
        of points in latitude/longitude coordinates.

        Parameters
        ----------

        points: numpy.ndarray of shape (n, 2) with dtype float
            The points to project on the globe given in latitude/longitude
            coordinates.

        Returns
        --------

        projection: numpy.ndarray of shape (n,2) or (n,3) with dtype float
            The points on the globe. The exact shape depends on the model.
        """

    @abstractmethod
    def lat_lon(self, points):
        """
        Computes the latitude/longitude representations of given points on the
        globe.

        Parameters
        ---------

        points: numpy.ndarray of shape (n,2) or (n,3) with dtype float
            The points on the globe. The exact shape depends on the model.

        Returns
        ---------

        lat_lons: numpy.ndarray of shape (n,2) with dtype float
            The latitude/longitude coordinates of the given points.
        """

    @abstractmethod
    def distance(self, start, end):
        """
        Computes the distance from `start` to `end` on the globe.

        Parameters
        ---------

        start: sequence of floats of length 2
            The latitude/longitude coordinates of the starting position.

        end: sequence of floats of length 2
            The latitude/longitude coordinates of the goal position.

        Returns
        -------

        distance: float
            The distance of the projections of the points `start` and `end`
            in the globe model.
        """

    @abstractmethod
    def shortest_projected_path(self, start, end, res=1000):
        """
        Computes the shortest path in the projected globe model.
        This is used in order to find the shortest path on the latitude-longitude plane.

        Parameters
        ----------

        start : np.ndarray of shape (2,n) or (3,n)
            The coordinates of the starting point on the globe.

        end : np.ndarray of shape (2,n) or (3,n)
            The coordinates of the destination point on the globe.

        res : int
            The number of points to be computed.

            Defaults to 1000.

        Returns
        ---------
        path: numpy.ndarray of shape (`res`, 2) or (`res`, 3)
            The coordinates of points along the shortest
            path from `start` to `end` on the projected globe.
        """

    def shortest_path(self, start, end, res=1000):
        """
        Uses the `shortest_projected_path` method to compute points on the
        shortest path from `start` to `end`, all given
        in latitude/longitude coordinates.

        Parameters
        ---------

        start : sequence of floats of length 2
            The latitude/longitude coordinates of the starting position.

        end : sequence of floats of length 2
            The latitude/longitude coordinates of the goal position.

        res : int
            The number of points to be computed.

            Defaults to 1000.

        Returns
        ----------

        path: numpy.ndarray of shape (`res`, 2)
            The latitude/longitude coordinates of points along the shortest
            path from `start` to `end`.

        """
        start, end = _ensure_2d(start, end)
        start, end = self.project(np.row_stack([start, end]))
        path = self.shortest_projected_path(start, end, res)
        return self.lat_lon(path)


class FlatMercatorProjection(GlobeModel):
    """
    Globe model that interprets coordinates via the mercator projection.

    Parameters
    ---------

    earth_radius: float/int
        The assumed radius of the (spherical) earth in nautical miles.

        Defaults to 3440.

    See also
    ----------
    `GlobeModel`
    """

    def __init__(
        self, virt_northpole=(90, 0), virt_zero=(0, 0), earth_radius=3440
    ):
        pole = _on_ball(np.array(virt_northpole))
        zero = _on_ball(np.array(virt_zero))
        east = np.cross(pole, zero)
        self._transform_mat = np.row_stack([zero, east, pole]).transpose()
        self._inverse_transform_mat = np.linalg.inv(self._transform_mat)

        self._earth_radius = earth_radius

    def _transform_coordinates(self, pts, matrix):
        pts = _on_ball(pts)
        pts = pts.reshape((len(pts), 3, 1))
        pts = matrix @ pts
        pts = pts.reshape(len(pts), 3)
        return _on_lat_lon(pts)

    def project(self, points):
        """
        Computes the mercator projection with reference point `lat_mp` of
        given points. Projection has size (n, 2).

        See also
        --------
        `GlobeModel.project`
        """
        points = np.array(points)
        points = self._transform_coordinates(points, self._transform_mat)
        lat, lon = points[:, 0], points[:, 1]

        return self._earth_radius * np.column_stack(
            [
                np.deg2rad(lon),
                np.arcsinh(np.tan(np.deg2rad(lat)))
                # np.log(np.tan(np.pi/4 + np.deg2rad(lat)/2))
            ]
        )

    def lat_lon(self, points):
        """
        See also
        ----------
        `GlobeModel.lat_lon`
        """
        points = np.array(points) / self._earth_radius
        x, y = points[:, 0], points[:, 1]
        lat = np.rad2deg(np.arcsin(np.tanh(y)))
        long = np.rad2deg(x)
        latlon = np.column_stack([lat, long])
        return self._transform_coordinates(latlon, self._inverse_transform_mat)

    def distance(self, start, end):
        """
        See also
        ---------
        `GlobeModel.distance`
        """
        start, end = _ensure_2d(start, end)
        return np.linalg.norm(self.project(start) - self.project(end))

    def shortest_projected_path(self, start, end, res=1000):
        """
        See also
        ---------
        `GlobeModel.shortest_projected_path`
        """
        return np.linspace(start, end, res)


class SphericalGlobe(GlobeModel):
    """
    Globe model that interprets latitude-longitude coordinates on a spherical
    globe.

    Parameters
    -----------
    earth_radius: int/float, optional
        The radius of the assumed globe.

    See also
    ----------
    `GlobeModel`
    """

    def __init__(self, earth_radius=21600 / 2 / np.pi):
        self._earth_radius = earth_radius

    def project(self, points):
        """
        See also
        ---------
        `GlobeModel.project`
        """
        return self._earth_radius * _on_ball(points)

    def lat_lon(self, points):
        """
        See also
        ---------
        `GlobeModel.lat_lon`
        """
        points = _ensure_2d(points)
        points = points / self._earth_radius
        return _on_lat_lon(points)

    def distance(self, start, end):
        """
        See also
        ---------
        `GlobeModel.distance`
        """

        start, end = _on_ball(np.row_stack([start, end]))
        angle = _angle(start, end)
        return self._earth_radius * angle

    def shortest_projected_path(self, start, end, res=1000):
        """
        See also
        ---------
        `GlobeModel.shortest_projected_path`
        """
        start, end = _ensure_2d(start, end)
        angle = _angle(start, end)

        angles = np.linspace(0, angle, res)

        # (1,0) -> (1, 0) -> (1,0,0) -> start, (cos(angle), sin(angle)) -> (0,1) -> (0,1,0) -> end
        proj1 = np.linalg.inv(
            np.array([[1, np.cos(angle)], [0, np.sin(angle)]])
        )
        proj2 = np.transpose(np.array([[1, 0, 0], [0, 1, 0]]))
        proj3 = np.transpose(np.row_stack([start, end, np.array([0, 0, 1])]))
        proj = proj3 @ proj2 @ proj1

        # # start -> (1,0,0) -> (1, 0) -> (1, 0), end -> (0,1,0) -> (0,1) -> (cos(angle), sin(angle))
        # inv_proj1 = np.linalg.inv(np.transpose(np.row_stack([start, end, np.array([0, 0, 1])])))
        # inv_proj2 = np.transpose(np.array([[1, 0], [0, 1], [0, 0]]))
        # inv_proj = np.linalg.inv(proj2)@inv_proj2@inv_proj1

        flat_pts = np.array([[[np.cos(ang)], [np.sin(ang)]] for ang in angles])
        return (proj @ flat_pts).reshape((len(flat_pts), 3))


def _ensure_2d(*args):
    if len(args) == 1:
        return np.atleast_2d(args[0])
    return (np.atleast_2d(pt) for pt in args)


def _on_ball(pts):
    pts = np.atleast_2d(np.deg2rad(pts))
    lat, lon = pts[:, 0], pts[:, 1]
    return np.column_stack(
        [np.cos(lat) * np.cos(lon), np.cos(lat) * np.sin(lon), np.sin(lat)]
    )


def _on_lat_lon(pts):
    x, y, z = pts.transpose()
    latlon = np.column_stack([np.arcsin(z), np.arctan2(y, x)])
    return np.rad2deg(latlon)


def _angle(pt1, pt2):
    """Computes angles in radians between points."""
    pt1 = pt1.ravel()
    pt2 = pt2.ravel()
    return np.arccos(
        np.dot(pt1, pt2) / np.linalg.norm(pt1) / np.linalg.norm(pt2)
    )
